"""
This package is a resource package containing various data files.

.. rubric:: XML Schema Files

* :file:`Ant-JUnit4.xsd`
* :file:`CTest-JUnit.xsd`
* :file:`Any-JUnit.xsd`
* :file:`GoogleTest-JUnit.xsd`
* :file:`PyTest-JUnit.xsd`
"""
